#! /usr/bin/env python

import unittest
import yaml
import psycopg2
import psycopg2.extras
from pirogue.merge import Merge, InvalidDefinition

pg_service = 'pirogue_test'


class TestMerge(unittest.TestCase):

    def setUp(self):
        self.conn = psycopg2.connect("service={0}".format(pg_service))
        self.cur = self.conn.cursor()

    def tearDown(self):
        self.conn.close()

    def test_insert_update_delete(self):
        yaml_definition = yaml.safe_load(open("test/merge_simple.yaml"))
        Merge(yaml_definition, pg_service=pg_service).create()

        # insert
        self.cur.execute("INSERT INTO pirogue_test.vw_merge_animal (animal_type,name,year,fk_cat_breed,eye_color) VALUES ('cat','felix',1985,2,'black');")
        self.cur.execute("SELECT animal_type, year, fk_cat_breed FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        res = self.cur.fetchone()
        self.assertEqual(res[0], 'cat')
        self.assertEqual(res[1], 1985)
        self.assertEqual(res[2], 2)
        # update
        self.cur.execute("SELECT * FROM pirogue_test.cat WHERE eye_color = 'black';")
        self.assertIsNotNone(self.cur.fetchone())
        self.cur.execute("UPDATE pirogue_test.vw_merge_animal SET animal_type = 'dog' WHERE name = 'felix';")
        self.cur.execute("SELECT * FROM pirogue_test.cat WHERE eye_color = 'black';")
        self.assertIsNone(self.cur.fetchone())
        # delete
        self.cur.execute("SELECT * FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.assertIsNotNone(self.cur.fetchone())
        self.cur.execute("DELETE FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.cur.execute("SELECT * FROM pirogue_test.vw_merge_animal WHERE name = 'felix';")
        self.assertIsNone(self.cur.fetchone())

    def test_type_change_not_allowed(self):
        yaml_definition = yaml.safe_load(open("test/merge_simple.yaml"))
        yaml_definition['view_name'] = 'vw_animal_no_type_change'
        yaml_definition['allow_type_change'] = False
        Merge(yaml_definition, pg_service=pg_service).create()
        self.cur.execute("INSERT INTO pirogue_test.vw_animal_no_type_change (animal_type,name,year,fk_cat_breed,eye_color) VALUES ('dog','albert',1933,2,'yellow');")
        error_caught = False
        try:
            self.cur.execute("UPDATE pirogue_test.vw_animal_no_type_change SET animal_type = 'cat';")
        except psycopg2.errors.RaiseException:
            error_caught = True
        self.assertTrue(error_caught)

    def test_invalid_definition(self):
        yaml_definition = yaml.safe_load(open("test/merge_simple.yaml"))
        yaml_definition['MyBadKey'] = 'Ouch'
        error_caught = False
        try:
            Merge(yaml_definition, pg_service=pg_service).create()
        except InvalidDefinition:
            error_caught = True
            self.assertTrue(error_caught)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

from psycopg2.extensions import cursor
from enum import Enum
from pirogue.information_schema import columns, primary_key, default_value


class InvalidColumn(Exception):
    pass


def table_parts(name: str) -> (str, str):
    """
    Returns a tuple with schema and table names
    :param name:
    :return:
    """
    if name and '.' in name:
        return name.split('.', 1)
    else:
        return 'public', name


def list2str(elements: list, sep: str= ', ', prepend: str='', append: str='', prepend_to_list: str='') -> str:
    """
    Prepend to all strings in the list
    :param elements:
    :param sep: separator
    :param prepend:
    :param append:
    :param prepend_to_list: prepend to the return string, if elements is not None or empty
    :return:
    """
    if elements is None or len(elements) == 0:
        return ''
    return prepend_to_list + sep.join([prepend+x+append for x in elements])


def column_alias(column: str,
                 remap_columns: dict = {},
                 prefix: str= None,
                 field_if_no_alias: bool = False,
                 prepend_as: bool = False) -> list:
    """

    :param table_alias:
    :param column:
    :param field_if_no_alias: if True, return the field if the alias doesn't exist. If False return an empty string
    :param prepend_as: prepend " AS " to the alias
    :return: empty string if there is no alias and (i.e = field name)
    """
    col_alias = ''
    if column in remap_columns:
        col_alias = remap_columns[column]
    elif prefix:
        col_alias = prefix + column
    elif field_if_no_alias:
        col_alias = column
    if prepend_as and col_alias:
        col_alias = ' AS {al}'.format(al=col_alias)
    return col_alias


def select_columns(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_type: str = 'table',
                   table_alias: str = None,
                   remove_pkey: bool = False,
                   skip_columns: list = [],
                   columns_list: list = None,
                   comment_skipped: bool = True,
                   remap_columns: dict = {},
                   columns_on_top: list = [],
                   columns_at_end: list = [],
                   prefix: str = None,
                   indent: int = 2) -> str:
    """

    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param table_alias: if not specified, table is used
    :param remove_pkey: if True, the primary is removed from the list
    :param skip_columns: list of columns to be skipped
    :param columns_list: if given use as list of columns
    :param comment_skipped: if True, skipped columns are written but commented, otherwise they are not written
    If remove_pkey is True, the primary key will not be printed
    :param remap_columns: dictionary to remap columns
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param indent: add an indent in front
    :return:
    """
    cols = sorted(columns_list or columns(pg_cur,
                                          table_schema=table_schema,
                                          table_name=table_name,
                                          table_type=table_type,
                                          remove_pkey=remove_pkey),
                  key=lambda col: __column_priority(col))

    # check arguments
    for param, dict_or_list in {'skip_columns': skip_columns,
                                'remap_columns': remap_columns,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))

    return ',\n{indent}'\
        .format(indent=indent*' ')\
        .join(['{skip}{table_alias}.{column}{col_alias}'
              .format(skip='-- ' if col in skip_columns else '',
                      table_alias=table_alias or table_name,
                      column=col,
                      col_alias=column_alias(col, remap_columns=remap_columns, prefix=prefix, prepend_as=True))
               for col in cols if (comment_skipped or col not in skip_columns)])


def insert_command(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_type: str = 'table',
                   table_alias: str = None,
                   remove_pkey: bool = True,
                   coalesce_pkey_default: bool = False,
                   skip_columns: list = [],
                   comment_skipped: bool = True,
                   remap_columns: dict = {},
                   insert_values: dict = {},
                   columns_on_top: list=[],
                   columns_at_end: list=[],
                   prefix: str = None,
                   returning: str = None,
                   indent: int = 2) -> str:
    """

    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param table_alias: the alias of the table
    :param remove_pkey: if True, the primary is removed from the list
    :param coalesce_pkey_default: if True, the following expression is used to insert the primary key: COALESCE( NEW.{pkey}, {default_value} )
    :param skip_columns: list of columns to be skipped
    :param comment_skipped: if True, skipped columns are written but commented, otherwise they are not written
    :param remap_columns: dictionary to remap columns
    :param insert_values: dictionary of expression to be used at insert
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param returning: returning command
    :param indent: add an indent in front
    :return:
    """
    # get columns
    cols = sorted(columns(pg_cur,
                          table_schema=table_schema,
                          table_name=table_name,
                          table_type=table_type,
                          remove_pkey=remove_pkey),
                  key=lambda col: __column_priority(col))

    pkey = None
    if coalesce_pkey_default:
        pkey = primary_key(pg_cur, table_schema, table_name)

    # check arguments
    for param, dict_or_list in {'skip_columns': skip_columns,
                                'remap_columns': remap_columns,
                                'insert_values': insert_values,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))

    def value(col):
        if col in insert_values:
            return '{val} -- {ori_col}'.format(val=insert_values[col], ori_col=col)
        cal = column_alias(col, remap_columns=remap_columns, prefix=prefix, field_if_no_alias=True)
        if coalesce_pkey_default and col == pkey:
            return 'COALESCE( NEW.{cal}, {pk_def} )'.format(cal=cal,
                                                            pk_def=default_value(pg_cur, table_schema, table_name, pkey))
        else:
            return 'NEW.{cal}'.format(cal=cal)

    next_comma_printed_1 = [False]
    next_comma_printed_2 = [False]
    return """INSERT INTO {s}.{t} (
{indent}      {cols} 
{indent}  ) VALUES ( 
{indent}      {new_cols}
{indent}  ){returning};
""".format(indent=indent*' ',
           s=table_schema,
           t=table_name,
           cols='\n{indent}    '
                .format(indent=indent*' ')
                .join(['{skip}{comma}{col}'
                      .format(indent=indent*' ',
                              skip='-- ' if col in skip_columns else '',
                              comma=', ' if __print_comma(next_comma_printed_1, col in skip_columns) else '',
                              col=col)
                       for col in cols if (comment_skipped or col not in skip_columns)]),
           new_cols='\n{indent}    '
                    .format(indent=indent*' ')
                    .join(['{skip}{comma}{value}'
                          .format(skip='-- ' if col in skip_columns else '',
                                  comma=', ' if __print_comma(next_comma_printed_2, col in skip_columns) else '',
                                  value=value(col))
                           for col in cols if (comment_skipped or col not in skip_columns)]),
           returning=' RETURNING {returning}'.format(indent=4*' ', returning=returning) if returning else '')


def update_command(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_alias: str = None,
                   table_type: str = 'table',
                   remove_pkey: bool = True,
                   pkey: str = None,
                   skip_columns: list=[],
                   comment_skipped: bool = True,
                   remap_columns: dict = {},
                   update_values: dict = {},
                   columns_on_top: list=[],
                   columns_at_end: list=[],
                   prefix: str= None,
                   where_clause: str = None,
                   indent: int=2) -> str:
    """
    Creates an UPDATE command
    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param remove_pkey: if True, the primary key will also be updated
    :param pkey: can be manually specified.
    :param table_alias: if not specified, table is used
    :param skip_columns: list of columns to be skipped
    :param comment_skipped: if True, skipped columns are written but commented, otherwise they are not written
    :param remap_columns: dictionary to remap columns
    :param update_values: dictionary of expression to be used at insert
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param where_clause: can be manually specified
    :param indent: add an indent in front
    :return: the SQL command
    """

    remove_pkey = remove_pkey and pkey is None and where_clause is None
    # get columns
    cols = sorted(columns(pg_cur,
                          table_schema=table_schema,
                          table_name=table_name,
                          table_type=table_type,
                          remove_pkey=remove_pkey),
                  key=lambda _col: __column_priority(_col))

    if pkey and remove_pkey:
        cols.remove(pkey)

    if not pkey and not where_clause:
        pkey = primary_key(pg_cur, table_schema, table_name)

    # check arguments
    for param, dict_or_list in {'skip_columns': skip_columns,
                                'remap_columns': remap_columns,
                                'update_values': update_values,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols and col != pkey:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))

    next_comma_printed = [False]

    return """UPDATE {s}.{t}{a} SET
{indent}    {cols}
{indent}  WHERE {where_clause};"""\
        .format(indent=indent*' ',
                s=table_schema,
                t=table_name,
                a=' {alias}'.format(alias=table_alias) if table_alias else '',
                cols='\n{indent}    '
                     .format(indent=indent*' ')
                     .join(['{skip}{comma}{col} = {new_col}'
                                .format(indent=indent*' ',
                                        skip='-- ' if col in skip_columns else '',
                                        comma=', ' if __print_comma(next_comma_printed, col in skip_columns) else '',
                                        col=col,
                                        new_col=update_values.get(col,
                                                                  'NEW.{cal}'.format(cal=column_alias(col,
                                                                                                      remap_columns=remap_columns,
                                                                                                      prefix=prefix,
                                                                                                      field_if_no_alias=True))))
                                for col in cols if (comment_skipped or col not in skip_columns)]),
                where_clause=where_clause or '{pkey} = {pkal}'.format(pkey=pkey,
                                                                      pkal=update_values.get(pkey,
                                                                                             'OLD.{cal}'.format(cal=column_alias(pkey,
                                                                                                                                 remap_columns=remap_columns,
                                                                                                                                 prefix=prefix,
                                                                                                                                 field_if_no_alias=True)))))


def __column_priority(column: str, columns_on_top: list=[], columns_at_end: list=[]) -> int:
    if column in columns_on_top:
        return 0
    elif column in columns_at_end:
        return 2
    else:
        return 1


def __print_comma(next_comma_printed: list, is_skipped: bool) -> bool:
    """
    Determines if a comma should be printed
    :param next_comma_printed: a list with a single boolean (works by reference)
    :param is_skipped:
    :return:
    """
    if is_skipped:
        return next_comma_printed[0]
    elif not next_comma_printed[0]:
        next_comma_printed[0] = True
        return False
    else:
        return True


# -*- coding: utf-8 -*-

import os
import psycopg2
import psycopg2.extras

from pirogue.utils import table_parts, column_alias, select_columns, insert_command, update_command
from pirogue.information_schema import TableHasNoPrimaryKey, NoReferenceFound, \
    columns, reference_columns, primary_key, default_value


class ReferencedTableDefinedBeforeReferencing(Exception):
    pass

class InvalidDefinition(Exception):
    pass

class VariableError(Exception):
    pass


class TableDef(object):
    table_schema = ''
    table_name = ''
    table_type = 'table'
    table_alias = None
    pkey = None  # determined automatically if not provided
    skip_columns = []
    comment_skipped = True
    remap_columns = {}
    prefix: str = None

    def __init__(self, table_schema: str, table_name: str):
        self.table_schema = table_schema
        self.table_name = table_name


class Merge:
    """
    Creates a merge view with associated triggers to edit.
    """

    def __init__(self, definition: dict, pg_service: str = None, variables: dict = {}):
        """
        Produces the SQL code of the join table and triggers
        :param definition: the YAML definition of the merge view
        :param pg_service:
        :param variables: dictionary for variables to be used in SQL deltas ( name => value )
        """

        self.variables = variables

        if pg_service is None:
            pg_service = os.getenv('PGSERVICE')
        self.conn = psycopg2.connect("service={0}".format(pg_service))
        self.cursor = self.conn.cursor()

        # check definition validity
        for key in definition.keys():
            if key not in ('table', 'view_schema',
                           'view_name', 'alias', 'short_alias',
                           'type_name', 'joins',
                           'insert_trigger', 'update_trigger',
                           'allow_type_change', 'allow_parent_only'):
                raise InvalidDefinition('key {k} is not a valid'.format(k=key))
        # check joins validity
        for alias, table_def in definition['joins'].items():
            for key in table_def.keys():
                if key not in ('table', 'short_alias',
                               'skip_columns', 'fkey',
                               'remap_columns', 'prefix', 'insert_values', 'update_values'):
                    raise InvalidDefinition('in join {a} key "{k}" is not valid'.format(a=alias, k=key))
        for mandatory_key in ['joins']:
            if mandatory_key not in definition:
                raise InvalidDefinition('Missing key: "{k}" should be provided.'.format(k=mandatory_key))
        # check trigger modifiers validity
        for trigger in ('insert_trigger', 'update_trigger'):
            for key in definition.get(trigger, {}):
                if key not in ('declare', 'pre', 'post'):
                    raise InvalidDefinition('key {k} is not valid in trigger definitions'.format(k=key))

        (self.master_schema, self.master_table) = table_parts(definition.get('table', None))
        self.master_skip_colums = definition.get('skip_columns', [])
        self.master_prefix = definition.get('prefix', None)
        self.master_remap_columns = definition.get('remap_columns', {})

        # global options:
        self.view_schema = definition.get('view_schema', self.master_schema)
        self.view_name = definition.get('view_name', "vw_merge_{t}".format(t=self.master_table))
        self.view_alias = definition.get('alias', self.master_table)
        self.short_alias = definition.get('short_alias', self.view_alias)
        self.type_name = definition.get('type_name', '{al}_type'.format(al=self.view_alias))
        self.insert_trigger = definition.get('insert_trigger', {})
        self.update_trigger = definition.get('update_trigger', {})
        self.allow_parent_only = definition.get('allow_parent_only', False)
        self.allow_type_change = definition.get('allow_type_change', True)

        try:
            self.master_pkey = primary_key(self.cursor, self.master_schema, self.master_table)
        except TableHasNoPrimaryKey:
            raise TableHasNoPrimaryKey('{vn} has no primary key, specify it with "key"'.format(vn=self.view_alias))

        # parse the joins definition
        self.joins = definition['joins']
        self.joined_ref_master_key = []
        for alias, table_def in self.joins.items():
            (table_def['table_schema'], table_def['table_name']) = table_parts(table_def['table'])
            table_def['short_alias'] = table_def.get('short_alias', alias)

            if 'fkey' in table_def:
                table_def['ref_master_key'] = table_def['fkey']
            else:
                table_def['ref_master_key'] = reference_columns(self.cursor,
                                                                table_def['table_schema'], table_def['table_name'],
                                                                self.master_schema, self.master_table)[0]
            try:
                table_def['pkey'] = primary_key(self.cursor, table_def['table_schema'], table_def['table_name'])
            except TableHasNoPrimaryKey:
                table_def['pkey'] = table_def['ref_master_key']

    def create(self) -> bool:
        """

        :return:
        """

        for sql in [self.__view(),
                    self.__insert_trigger(),
                    self.__update_trigger(),
                    self.__delete_trigger()]:
            try:
                if self.variables:
                    self.cursor.execute(sql, self.variables)
                else:
                    self.cursor.execute(sql)
            except TypeError as e:
                print("*** Failing:\n{}\n***".format(sql))
                raise VariableError("An error in a SQL variable is probable. "
                                    "Check the variables in the SQL code "
                                    "(were given: {svars}). "
                                    "Also, any % character shall be escaped with %%"
                                    .format(svars=list(self.variables.keys())))
            except psycopg2.Error as e:
                print("*** Failing:\n{}\n***".format(sql))
                raise e
        self.conn.commit()
        self.conn.close()
        return True

    def __view(self) -> str:
        """
        :return:
        """

        sql = """
CREATE OR REPLACE VIEW {vs}.{vn} AS
  SELECT
    CASE
      {types}
      ELSE {no_subtype}
    END AS {type_name},
    {master_columns},
    {joined_columns}
  FROM {mt}.{ms} {sa}
    {joined_tables};        
""".format(vs=self.view_schema,
           vn=self.view_name,
           types='\n      '.join(["WHEN {shal}.{mrf} IS NOT NULL THEN '{al}'::text"
                                 .format(shal=table_def['short_alias'], mrf=table_def['ref_master_key'], al=alias)
                                  for alias, table_def in self.joins.items()]),
           no_subtype="'unknown'::text",
           type_name=self.type_name,
           master_columns=select_columns(self.cursor, self.master_schema, self.master_table,
                                         table_alias=self.view_alias,
                                         skip_columns=self.master_skip_colums,
                                         prefix=self.master_prefix,
                                         remap_columns=self.master_remap_columns,
                                         indent=4),
           joined_columns=', '.join([select_columns(self.cursor, table_def['table_schema'], table_def['table_name'],
                                                    table_alias=table_def['short_alias'],
                                                    skip_columns=table_def.get('skip_columns', [])+[table_def['ref_master_key']],
                                                    prefix=table_def.get('prefix', None),
                                                    remove_pkey=False,
                                                    remap_columns=table_def.get('remap_columns', {}),
                                                    indent=4)
                                     for alias, table_def in self.joins.items()]),
           mt=self.master_schema,
           ms=self.master_table,
           sa=self.short_alias,
           joined_tables='\n    '.join(["LEFT JOIN {tbl} {tal} ON {tal}.{rmk} = {msa}.{mpk}"
                                            .format(tbl=table_def['table'],
                                                    tal=table_def['short_alias'],
                                                    rmk=table_def['ref_master_key'],
                                                    msa=self.short_alias,
                                                    mpk=self.master_pkey)
                                            for table_def in self.joins.values()])
            )
        return sql

    def __insert_trigger(self) -> str:
        """

        :return:
        """
        sql = """-- INSERT TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_insert() RETURNS trigger AS
$BODY$
DECLARE
  {declare}
BEGIN
  {insert_trigger_pre}
  {insert_master}

  CASE 
    {insert_joins}
  ELSE
    {raise_notice}
  END CASE;

  {insert_trigger_post}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_insert ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_insert
    INSTEAD OF INSERT ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_insert();
""".format(vs=self.view_schema,
           vn=self.view_name,
           declare='\n  '.join(['{d};'.format(d=declare) for declare in self.insert_trigger.get('declare', [])]),
           insert_trigger_pre=self.insert_trigger.get('pre', ''),
           insert_master=insert_command(self.cursor, self.master_schema, self.master_table,
                                        skip_columns=self.master_skip_colums,
                                        prefix=self.master_prefix,
                                        remap_columns=self.master_remap_columns,
                                        remove_pkey=False,
                                        indent=8,
                                        coalesce_pkey_default=True,
                                        returning='{mpk} INTO NEW.{mpk}'.format(mpk=self.master_pkey)),
           insert_joins='\n    '.join(["WHEN NEW.{type_name} = '{alias}' THEN"
                                       "\n      {insert_join}".format(type_name=self.type_name,
                                                                      alias=alias,
                                                                      insert_join=insert_command(self.cursor,
                                                                                                 table_def['table_schema'],
                                                                                                 table_def['table_name'],
                                                                                                 table_alias=table_def['short_alias'],
                                                                                                 skip_columns=table_def.get('skip_columns', []),
                                                                                                 prefix=table_def.get('prefix', None),
                                                                                                 insert_values={**{table_def['ref_master_key']: 'NEW.{c}'.format(c=self.master_pkey)},
                                                                                                                **table_def.get('insert_values', {})},
                                                                                                 remap_columns=table_def.get('remap_columns', {}),
                                                                                                 remove_pkey=False,
                                                                                                 indent=4))
                                   for alias, table_def in self.joins.items()]),
           raise_notice='NULL;' if self.allow_parent_only else "RAISE NOTICE '{vn} type not known ({percent_char})', NEW.{type_name}; -- ERROR"
                            .format(vn=self.view_name,
                                    percent_char='%%' if self.variables else '%',  # if variables, % should be escaped because cursor.execute is run with variables
                                    type_name=self.type_name),
           insert_trigger_post=self.insert_trigger.get('post', ''))
        return sql


    def __update_trigger(self):
        sql = """-- UPDATE TRIGGER
CREATE OR REPLACE FUNCTION {vs}.ft_{vn}_update() RETURNS trigger AS
$BODY$
DECLARE
  {declare}
BEGIN
  {update_trigger_pre}
  {update_master}

  IF OLD.{type_name} <> NEW.{type_name} THEN
    {type_change}
  END IF;

  CASE {update_joins}
  ELSE
     {raise_notice}
  END CASE;
  {update_trigger_post}
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

DROP TRIGGER IF EXISTS tr_{vn}_on_udpate ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_update
    INSTEAD OF update ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_update();
        """.format(vs=self.view_schema,
                   vn=self.view_name,
                   declare='\n  '.join(['{d};'.format(d=declare) for declare in self.update_trigger.get('declare', [])]),
                   update_trigger_pre=self.update_trigger.get('pre', ''),
                   update_master=update_command(self.cursor, self.master_schema, self.master_table,
                                                skip_columns=self.master_skip_colums,
                                                prefix=self.master_prefix,
                                                remap_columns=self.master_remap_columns,
                                                indent=8),
                   type_name=self.type_name,
                   type_change="RAISE EXCEPTION 'Type change not allowed for {alias}'"
                               "\n      USING HINT = 'You cannot switch from ' "
                               "|| OLD.{type_name} || ' to ' || NEW.{type_name};"
                                .format(alias=self.view_alias, type_name=self.type_name)
                               if not self.allow_type_change else
                               "CASE"
                               "\n      {deletes}"
                               "\n    END CASE;"
                               "\n    CASE"
                               "\n      {inserts}"
                               "\n      ELSE -- do nothing"
                               "\n    END CASE;"
                               .format(deletes='\n      '.join(["WHEN OLD.{type_name} = '{alias}' "
                                                                "THEN DELETE FROM {ts}.{tn} "
                                                                "WHERE {rmk} = OLD.{mpk};"
                                                                .format(type_name=self.type_name,
                                                                        alias=alias,
                                                                        ts=table_def['table_schema'],
                                                                        tn=table_def['table_name'],
                                                                        rmk=table_def['ref_master_key'],
                                                                        mpk=self.master_pkey)
                                                                for alias, table_def in self.joins.items()]),
                                       inserts='\n      '.join(["WHEN NEW.{type_name} = '{alias}' "
                                                                "THEN INSERT INTO {ts}.{tn} "
                                                                "({rmk}) VALUES (OLD.{mpk});"
                                                               .format(type_name=self.type_name,
                                                                       alias=alias,
                                                                       ts=table_def['table_schema'],
                                                                       tn=table_def['table_name'],
                                                                       rmk=table_def['ref_master_key'],
                                                                       mpk=self.master_pkey)
                                                                for alias, table_def in self.joins.items()])),
                   update_joins='\n    '.join(["WHEN NEW.{type_name} = '{alias}' THEN"
                                               "\n      {update_join}".format(type_name=self.type_name,
                                                                              alias=alias,
                                                                              update_join=update_command(self.cursor, table_def['table_schema'], table_def['table_name'],
                                                                                                         table_alias=table_def['short_alias'],
                                                                                                         pkey=table_def['pkey'],
                                                                                                         skip_columns=table_def.get('skip_columns', []),
                                                                                                         prefix=table_def.get('prefix', None),
                                                                                                         remap_columns=table_def.get('remap_columns', {}),
                                                                                                         update_values={**{table_def['pkey']: 'OLD.{c}'.format(c=self.master_pkey)},
                                                                                                                        **table_def.get('update_values', {})},
                                                                                                         indent=4))
                                   for alias, table_def in self.joins.items()]),
                   raise_notice='NULL;' if self.allow_parent_only
                                else "RAISE NOTICE '{vn} type not known ({percent_char})', NEW.{type_name}; -- ERROR"
                                     .format(vn=self.view_name,
                                             percent_char='%%' if self.variables else '%',  # if variables, % should be escaped because cursor.execute is run with variables
                                             type_name=self.type_name),
                   update_trigger_post=self.update_trigger.get('post', ''))
        return sql

    def __delete_trigger(self):
        sql = """
CREATE FUNCTION {vs}.ft_{vn}_delete() RETURNS trigger AS
    $BODY$
    BEGIN
    CASE
        {deletes}
    END CASE;
    DELETE FROM {ts}.{tn} WHERE {mpk} = OLD.{mpk};
    RETURN NULL;
    END;
    $BODY$
    LANGUAGE 'plpgsql';

DROP TRIGGER IF EXISTS tr_{vn}_on_delete ON {vs}.{vn};

CREATE TRIGGER tr_{vn}_on_delete
    INSTEAD OF DELETE ON {vs}.{vn}
    FOR EACH ROW EXECUTE PROCEDURE {vs}.ft_{vn}_delete();
""".format(vn=self.view_name,
           deletes='\n      '.join(["WHEN OLD.{type_name} = '{alias}' "
                                    "THEN DELETE FROM {ts}.{tn} "
                                    "WHERE {rmk} = OLD.{mpk};"
                                   .format(type_name=self.type_name,
                                           alias=alias,
                                           ts=table_def['table_schema'],
                                           tn=table_def['table_name'],
                                           rmk=table_def['ref_master_key'],
                                           mpk=self.master_pkey)
                                    for alias, table_def in self.joins.items()]),
           ts=self.master_schema,
           tn=self.master_table,
           mpk=self.master_pkey,
           vs=self.view_schema)
        return sql